#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "my_types.h"
#include "oggStream.h"
#include "codebook.h"
#include "floor1.h"
#include "bitset.h"
#include "residue.h"
#include "vorbis_defs.h"
#include "helpers.h"

int
decode_lookup_table1(codebook_t *cb)
{
dbg_printf(stderr, "Decoding lookup table type 1.\n");
	UINT32 lookup_offset;
	cb->value_vectors = (FIXINT32**)my_alloc(cb->num_entries*sizeof(FIXINT32*));
	if (cb->value_vectors == NULL) {
		err_printf(stderr, "Error allocating value_vectors array.\n");
		return -1;
	}
	for(lookup_offset=0; lookup_offset < cb->num_entries; lookup_offset++) {
		UINT32 index_divisor = 1;
		FIXINT32 last = 0.0F;
		UINT16 i = 0;
		cb->value_vectors[lookup_offset] = 
			(FIXINT32*)my_alloc(cb->dimensions*sizeof(FIXINT32));
		if (cb->value_vectors[lookup_offset] == NULL) {
			err_printf(stderr, "Error allocating value vector %u.\n",
					lookup_offset);
			return -1;
		}
		for(i = 0; i < cb->dimensions; i++) {
			UINT32 mult_off = 
					(UINT32)(lookup_offset/index_divisor)
					% cb->lookup_values;
			cb->value_vectors[lookup_offset][i] =
				cb->multiplicands[mult_off] * cb->delta_val
				+ cb->min_val + last;
			if (cb->sequence_p) {
				last = cb->value_vectors[lookup_offset][i];
			}
			index_divisor = index_divisor * cb->lookup_values;
		}
	}
	return 0;
}

int
decode_lookup_table2(codebook_t *cb)
{
dbg_printf(stderr, "Decoding lookup table type 2.\n");
	UINT32 lookup_offset;
	cb->value_vectors = (FIXINT32**)my_alloc(cb->num_entries*sizeof(FIXINT32*));
	if (cb->value_vectors == NULL) {
		err_printf(stderr, "Error allocating value_vectors array.\n");
		return -1;
	}
	for(lookup_offset=0; lookup_offset < cb->num_entries; lookup_offset++) {
		FIXINT32 last = 0.0F;
		UINT16 i = 0;
		UINT32 mult_off = lookup_offset * cb->dimensions;
		cb->value_vectors[lookup_offset] = 
			(FIXINT32*)my_alloc(cb->dimensions*sizeof(FIXINT32));
		if (cb->value_vectors[lookup_offset] == NULL) {
			err_printf(stderr, "Error allocating value vector %u.\n",
					lookup_offset);
			return -1;
		}
		for(i = 0; i < cb->dimensions; i++) {
			cb->value_vectors[lookup_offset][i] =
				cb->multiplicands[mult_off] * cb->delta_val
				+ cb->min_val + last;
			if (cb->sequence_p) {
				last = cb->value_vectors[lookup_offset][i];
			}
			mult_off++;
		}
	}
	return 0;
}

int
decode_lookup_table(codebook_t *cb)
{
	switch(cb->lookup_type) {
	case 0:	// Type 0 means no lookup table
		return 0;
	case 1:
		return decode_lookup_table1(cb);
	case 2:
		return decode_lookup_table2(cb);
	default: // Everything except 0,1, and 2 is reserved => error
		return -1;
	}
	// If we ever get here it's an error
	return -1;
}

#define NEW_INTERNAL_NODE(n)						\
	h_node_t *n = (h_node_t*)my_alloc(sizeof(h_node_t));		\
	n->value = INTERNAL_NODE_VALUE;					\
	n->child[0] = n->child[1] = NULL;

h_node_t*
find_next_node(UINT32 length, h_node_t* root)
{
	if (root == NULL || root->value != INTERNAL_NODE_VALUE) {
		return NULL;
	}
	if (length == 0) {
		if ( (root->value == INTERNAL_NODE_VALUE)
				&& (root->child[0] == NULL)
				&& (root->child[1] == NULL) ) {
			return root;
		} else {
			return NULL;
		}
	} 
	if (root->child[0] == NULL) {
		NEW_INTERNAL_NODE(child);
		root->child[0] = child;
		return find_next_node(length-1, child);
	} else {
		h_node_t *node = find_next_node(length-1, root->child[0]);
		if (node != NULL) {
			return node;
		}
		if (root->child[1] == NULL) {
			NEW_INTERNAL_NODE(child);
			root->child[1] = child;
			return find_next_node(length-1, child);
		} else {
			return find_next_node(length-1, root->child[1]);
		}
	}
	// We shouldn't get here, but return NULL just in case.
	return NULL;
}

void
print_tree(h_node_t *root, UINT32 codeword, UINT32 length)
{
	if (root->value != INTERNAL_NODE_VALUE) {
		UINT32 i = length;
		for(; i > 0; i--) {
			err_printf(stderr, "%01u", 
					((codeword & (1<<(i-1))) >> (i-1)));
		}
		err_printf(stderr, " -> %u (%p-%u)\n", root->value, root, length);
		return;
	} else {
		if (root->child[0] != NULL) {
			print_tree(root->child[0], (codeword << 1), (length+1));
		}
		if (root->child[1] != NULL) {
			print_tree(root->child[1], 
					((codeword << 1) | 1), (length+1));
		}
	}
}

void
print_tree_node(h_node_t *root, h_node_t *node, UINT32 codeword, UINT32 length)
{
	if (root == node) {
		err_printf(stderr, "0%o\n", codeword);
		return;
	} else if (root->value != INTERNAL_NODE_VALUE) {
		return;
	} else {
		if (root->child[0] != NULL) {
			print_tree_node(root->child[0], node,
					(codeword << 1), (length+1));
		}
		if (root->child[1] != NULL) {
			print_tree_node(root->child[1], node,
					((codeword << 1) | 1), (length+1));
		}
	}
}

void
printCodeword(h_node_t* root, h_node_t* node)
{
	err_printf(stderr, ", codeword = ");
	print_tree_node(root, node, 0, 0);
}

/*
 * Build a Huffman decision tree for the code book.
 * The values in cb->lengths determine the length of the code words.
 * The lowest available codeword of the correct length is chose.
 * The actual huffman tree only contains child pointers to save space,
 * so a stack is used to remeber ancestors while building the tree.
 */

int
build_huffman_tree(codebook_t *cb)
{
	NEW_INTERNAL_NODE(root);
	cb->huffman = root;
	register UINT32 i = 0;
	register UINT32 entry = 0;
	for(; i < cb->num_entries; i++) {
		UINT32 length = cb->lengths[i];
		h_node_t *node;

		if (cb->lengths[i] == 0) {
			if (cb->sparse) entry++;
			continue;
		}

		node = find_next_node(length, root);
		if (node == NULL) {
			err_printf(stderr, "Failed to find an available node "
					" of correct length.\n");
			return -1;
		} else {
//dbg_printf(stderr, "Added codeword of length %d value %d", length, entry);
#ifdef VERBOSE_OUTPUT
//printCodeword(root, node);
#endif
			node->value = entry++;
		}
	}
#if 0
dbg_printf(stderr, "Printing codebook rooted at %p\n", root);
print_tree(root, 0, 0);
#endif
	return 0;
}

UINT32
read_codeword(OggStream &ogg, codebook_t &cb)
{
	h_node_t *node = cb.huffman;
//dbg_printf(stderr, "Reading codeword from codebook rooted at %p\n", node);
int codeword = 0;
int len = 0;

	while(node != NULL && node->value == INTERNAL_NODE_VALUE) {
		int child = ogg.readBits(1);
		if (child < 0) {
dbg_printf(stderr, "Failed to read bit %d, returning 0xFFFFFFFF\n", len+1);
			return 0xffffffff;
		}
assert((child == 0) || (child == 1));
codeword <<= 1;
codeword |= (child & 0x1);
len++;
		node = node->child[child];
	}
#if 0
dbg_printf(stderr, "Read codeword ");
for(int i = len; i > 0; i--) {
	err_printf(stderr, "%01u", ((codeword & (1<<(i-1))) >> (i-1)));
}
if (node == NULL) {
	err_printf(stderr, " and node was null.\n");
} else {
	err_printf(stderr, " and value was %u.\n", node->value);
}
#endif
	if (node == NULL) return 0xffffffff;
//dbg_printf(stderr, "Read codeword 0%o (val = %d)\n", codeword, node->value);
	return node->value;
}

FIXINT32*
read_code_vector(OggStream &ogg, codebook_t &cb)
{
//dbg_printf(stderr, "Reading vector from codebook at %p\n", &cb);
	if (cb.lookup_type == 0) {
dbg_printf(stderr, "trying to read vector from codebook with lookup type 0.\n");
		return NULL;
	}
	UINT32 value = read_codeword(ogg, cb);
	if (value > cb.num_entries || value == 0xFFFFFFFF) {
dbg_printf(stderr, "value = %u, n_entries = %u, returning NULL\n",
value, cb.num_entries);
		return NULL;
	}
//dbg_printf(stderr, "Read codeword vector at offset %d\n", value);
	return cb.value_vectors[value];
}

void
free_tree(h_node_t *root)
{
	if (root == NULL) return;

	free_tree(root->child[0]);
	free_tree(root->child[1]);
	free(root);
}
