#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <assert.h>

using namespace std;

#include "my_types.h"
#include "helpers.h"
#include "mdct.h"

static int count = -1;

int blocksizes[2] = { 0 };
FIXINT32 *trig[2][3] = { { NULL, NULL, NULL}, { NULL, NULL, NULL} };

int
init_MDCT(int blocksize_0, int blocksize_1)
{
	clean_MDCT();	// Clean-up just in case

	blocksizes[0] = blocksize_0;
	blocksizes[1] = blocksize_1;
	// Allocate trig lookups
	for(int i = 0; i < 2; i++) {
		int size = blocksizes[i] >> 1;
		trig[i][2] = (FIXINT32*)my_alloc(size*sizeof(FIXINT32));
		if (trig[i][2] == NULL) {
			goto init_err;
		}
		size >>= 2;
		trig[i][0] = (FIXINT32*)my_alloc((size+1)*sizeof(FIXINT32));
		if (trig[i][0] == NULL) {
			goto init_err;
		}
		trig[i][1] = (FIXINT32*)my_alloc(size*sizeof(FIXINT32));
		if (trig[i][1] == NULL) {
			goto init_err;
		}
	}

	// Populate the lookup tables.

	// The first table is the values for cos(4*k*pi/N) and sin(4*k*pi/N)
	// We only store one quarter of the sinusoid to save space
	for(int i = 0; i < 2; i++) {
		int N = blocksizes[i];
		FIXINT32 *T = trig[i][0];
		*(T++) = 1.0f;  // Hard code the first value for added precision
		for(int k = 1; k < (N>>3); k++,T++) {
			*T = (float)cos(4*k*pi/(FLOAT32)N);
		}
		*T = 0.0f; // Hard code the last value for added precision
	}
	
	// The second table is the values for cos(2(2k+1)*pi/N) 
	// and sin(2(2k+1)*pi/N)
	// the sine values are the same as the cosine values in reverse order
	// We store all values for the range k = 0,1,..,N/8-1
	for(int i = 0; i < 2; i++) {
		int N = blocksizes[i];
		FIXINT32 *T = trig[i][1];
		for(int k = 0; k < (N>>3); k++,T++) {
			*T = (float)cos(2*(2*k+1)*pi/(FLOAT32)N);
		}
	}
	
	// The third table is the values for cos((2k+1)*pi/(2N)) 
	// and sin((2k+1)*pi/(2N))
	// We store all values for the range k = 0,1,..,N/8-1
	// We interleave the cosines and sines because they're always used
	// in pairs
	for(int i = 0; i < 2; i++) {
		int N = blocksizes[i];
		FIXINT32 *T = trig[i][2];
		for(int k = 0; k < (N>>2); k++,T+=2) {
			T[0] = (float)cos((2*k+1)*pi/(FLOAT32)(N<<1));
			T[1] = (float)sin((2*k+1)*pi/(FLOAT32)(N<<1));
		}
	}

	return 0;

init_err:
	clean_MDCT();
	return -1;
}

void
clean_MDCT()
{
	for(int i = 0; i < 2; i++) {
		for (int j = 0; j < 3; j++) {
			if (trig[i][j] != NULL) {
				free(trig[i][j]);
				trig[i][j] = NULL;
			}
		}
		blocksizes[i] = 0;
	}
}

static inline void
iMDCT_step_1(FIXINT32 *data, int n2, int n4, FIXINT32* T)
{
	FIXINT32 *in = data + n2;
	FIXINT32 *out1 = data + n2 + n4;
	FIXINT32 *out2 = data + n2 + n4;
	
	in -= 4;
	out1 -= 2;
	FIXINT32 *c1 = T + (n4>>1);
	FIXINT32 *c2 = T + (n4>>1) - 1;
	FIXINT32 *s1 = T;
	FIXINT32 *s2 = T + 1;

	for(int k = (n4>>1); k < n4; k++,c1--,c2--,s1++,s2++) {
		
//assert((out1 >= data) && ((out1+2) <= (data+(n2<<1))));
//assert((out2 >= data) && ((out2+2) <= (data+(n2<<1))));
//assert((in >= data) && ((in+4) <= (data+(n2<<1))));
		out1[0] = in[3]*(*s1)  + in[1]*(*c1);
		out1[1] = in[3]*(*c1) - in[1]*(*s1);
		out2[1] = in[0]*(*c2) + in[2]*(*s2);
		out2[0] = in[0]*(s2->negate()) + in[2]*(*c2); 
		out1 -= 2;
		out2 += 2;
		in -= 4;
	}
}

static inline void
iMDCT_step_2(FIXINT32 *data, int n2, int n4, int N, FIXINT32 *T)
{
	FIXINT32 *d1 = data + n2 + n4;
	FIXINT32 *d2 = data + n2;
	FIXINT32 *d3 = data + N - 2;
	FIXINT32 *d4 = data + n2 + n4 - 2;

	FIXINT32 *c = T;
	FIXINT32 *s = T + (n4>>1) - 1;

	int k = (n4 - 2);
	for(; k >= (n4>>1); k -= 2,d1+=2,d2+=2,d3-=2,d4-=2,c++,s--) {
		register FIXINT32 r0 = d1[1] - d2[1];
		register FIXINT32 r1 = d1[0] - d2[0];
//assert((d1 >= data) && ((d1+2) <= (data+N)));
//assert((d2 >= data) && ((d2+2) <= (data+N)));

		d1[1] += d2[1];
		d1[0] += d2[0];
		d2[1] =  r1*(s[1]) - r0*(c[0]);
		d2[0] =  r0*(s[1].negate()) - r1*(c[0]);
		
		register FIXINT32 r2 = d3[1] - d4[1];
		register FIXINT32 r3 = d3[0] - d4[0];
//assert((d3 >= data) && ((d3+2) <= (data+N)));
//assert((d4 >= data) && ((d4+2) <= (data+N)));

		d3[1] += d4[1];
		d3[0] += d4[0];
		d4[1] =   r2*(c[1]) + r3*(s[0]);
		d4[0] =  r3*(c[1]) - r2*(s[0]);
	}
}

// The butterfly step
static inline void
iMDCT_butterfly(FIXINT32 *data, int n2, int n4, int N, FIXINT32 *T)
{
	UINT32 stride = n4;
	UINT32 limit = 2;
	for(; stride >= 4; stride >>= 1, limit <<= 1) {
		for(UINT32 s = 0; s < limit; s++) {
			FIXINT32 *o1 = data+N-1-(s*stride);
			FIXINT32 *o2 = o1 - 1;
			FIXINT32 *o3 = o1 - (stride>>1);
			FIXINT32 *o4 = o3 - 1;
			if ((stride>>2) == 1) {
				register FIXINT32 r0 = *o1 - *o3;
				register FIXINT32 r1 = *o2 - *o4;
//assert((o1 >= data) && (o1 < (data+N)));
//assert((o2 >= data) && (o2 < (data+N)));
//assert((o3 >= data) && (o3 < (data+N)));
//assert((o4 >= data) && (o4 < (data+N)));
				*o1 += *o3;
				*o2 += *o4;
				*o3 = r0;
				*o4 = r1;
				continue;
			}
			unsigned r = 0;
			unsigned t_off = (limit<<1);
			FIXINT32 *c = T;
			FIXINT32 *s = T + (n4>>1);
			for(; r < (stride>>3); o1-=2,o2-=2,o3-=2,o4-=2,r++){
				FIXINT32 A_rk1 = (*c);
				FIXINT32 A_rk1_1 = (*s).negate();
				register FIXINT32 r0 = *o1 - *o3;
				register FIXINT32 r1 = *o2 - *o4;
//assert((o1 >= data) && (o1 < (data+N)));
//assert((o2 >= data) && (o2 < (data+N)));
//assert((o3 >= data) && (o3 < (data+N)));
//assert((o4 >= data) && (o4 < (data+N)));
				*o1 += *o3;
				*o2 += *o4;
				*o3 = r0*(A_rk1) - r1*(A_rk1_1);
				*o4 = r1*(A_rk1) + r0*(A_rk1_1);
				// Update Trig lookup pointers
				c += t_off;
				s -= t_off;
			}
			c = T + (n4>>1);
			s = T;
			for(; r < (stride>>2); o1-=2,o2-=2,o3-=2,o4-=2,r++){
				FIXINT32 A_rk1 = ((*c).negate());
				FIXINT32 A_rk1_1 = ((*s).negate());
				register FIXINT32 r0 = *o1 - *o3;
				register FIXINT32 r1 = *o2 - *o4;
//assert((o1 >= data) && (o1 < (data+N)));
//assert((o2 >= data) && (o2 < (data+N)));
//assert((o3 >= data) && (o3 < (data+N)));
//assert((o4 >= data) && (o4 < (data+N)));
				*o1 += *o3;
				*o2 += *o4;
				*o3 = r0*(A_rk1) - r1*(A_rk1_1);
				*o4 = r1*(A_rk1) + r0*(A_rk1_1);
				c -= t_off;
				s += t_off;
			}
		}
	}
}

static inline int
bitreverse(int in, int bits)
{
	int out = 0;
	for(int i = 0; i < bits; i++,in>>=1) {
		out <<=1;
		out |= (in & 0x1);
	}
	return out;
}

static inline void
iMDCT_bitreverse(FIXINT32 *data, int N, int n_bits)
{
	FIXINT32 *start = data + (N/2);
	FIXINT32 *i_data = data + (N/2) + 4;
	for(int i = 1; i < ((N/8)-2); i++, i_data+=4) {
		int j = bitreverse(i,n_bits);
		if (i < j) {
			FIXINT32 *j_data = start + (4*j);
			for(int k = 0; k < 4; k++) {
//assert((i_data >= data) && ((i_data+4) <= (data+N)));
//assert((j_data >= data) && ((j_data+4) <= (data+N)));
				register FIXINT32 temp = i_data[k];
				i_data[k] = j_data[k];
				j_data[k] = temp;
			}
		}
	}
}

static inline void
iMDCT_step_7(FIXINT32 *data, int N, int n2, int n4, int n8, FIXINT32* T)
{
	FIXINT32 *x1 = data + n2;
	FIXINT32 *x2 = data + N - 2;
	FIXINT32 *c = T;
	FIXINT32 *s = T + n8 - 1;

	for(int k = 0; k < n4; k+=2,x1+=4,x2-=4,c++,s--) {
//assert((x1 >= data) && ((x1+2) <= (data+N)));
//assert((x2 >= data) && ((x2+2) <= (data+N)));
		register FIXINT32 r0 = x1[0] + x2[0]; // b+d
		register FIXINT32  r1 = x1[1] + x2[1]; // a+c
		register FIXINT32 r2 = x2[0] - x1[0]; // b-d
		register FIXINT32 r3 = x2[1] - x1[1]; // a-c
		x2[1] = (r1 + r3*(s->negate()) + r0*(*c)).div2();
		x1[1] = (r1 - r3*(s->negate()) - r0*(*c)).div2();
		x2[0] = (r2 + r0*(s->negate()) - r3*(*c)).div2();
		x1[0] = (r0*(s->negate()) - r2 - r3*(*c)).div2();
	}
}

static inline void
iMDCT_step_8_1(FIXINT32 *data, int N, int n4, int n8, FIXINT32 *T)
{
	FIXINT32 *in1 = data + N - 4;
	FIXINT32 *out1 = data + n4 - 1;
	FIXINT32 *out2 = data + n4;
	FIXINT32 *out3 = data + n8 - 1;
	FIXINT32 *out4 = data + 3*n8;

	FIXINT32 *t1 = T;
	FIXINT32 *t2 = T + n4;
	for(int k = 0; k < n4; k+=2,in1-=4,out1--,out2++,out3--,out4++){
//assert((in1 >= data) && ((in1+2) <= (data+N)));
//assert((out1 >= data) && ((out1+1) <= (data+N)));
//assert((out2 >= data) && ((out2+1) <= (data+N)));
//assert((out3 >= data) && ((out3+1) <= (data+N)));
//assert((out3 >= data) && ((out3+1) <= (data+N)));
		*out1 = in1[3]*t1[1] - in1[2]*t1[0];
		*out2 = in1[3]*t1[0] + in1[2]*t1[1];
		*out3 = in1[1]*t2[1] - in1[0]*t2[0];
		*out4 = in1[1]*t2[0] + in1[0]*t2[1];
	}
}

static inline void
iMDCT_step_8_2(FIXINT32 *data, int N)
{	
	FIXINT32 *in1 = data;
	FIXINT32 *out1 = data +  N/2;
	FIXINT32 *out2 = data + N - 1;
	FIXINT32 *in2 = out1 - 1;
	for(int k = 0; k < (N/4); k++,in1++,in2--,out1++,out2--) {
//assert((out1 >= data) && ((out1+1) <= (data+N)));
//assert((out2 >= data) && ((out2+1) <= (data+N)));
//assert((in1 >= data) && ((in1+1) <= (data+N)));
//assert((in2 >= data) && ((in2+1) <= (data+N)));
		*out1 = (*in2).negate();
		*out2 = (*in2).negate();
		*in2 =  (*in1).negate();
	}
}

void printData(FLOAT32 *data, int N, const char* fname);

void
inverse_MDCT(FIXINT32 *data_int, int N, int n_bits)
{
	
  count++;
	int n2 = N >> 1;
	int n4 = N >> 2;
	int n8 = N >> 3;
	int b;
	if (N == blocksizes[0]) {
		b = 0;
	} else if (N == blocksizes[1]) {
		b = 1;
	} else {
		return;
	}
	
	iMDCT_step_1(data_int, n2, n4, trig[b][0]);

	iMDCT_step_2(data_int, n2, n4, N, trig[b][0]);

	iMDCT_butterfly(data_int, n2, n4, N, trig[b][0]);
	
	iMDCT_bitreverse(data_int, N, n_bits);
	
	// Steps 5 and 6 just mess around with the indices
	// We know what the corrected indices are
	// So we can just correct the indices in steps 7 and 8
	iMDCT_step_7(data_int, N, n2, n4, n8, trig[b][1]);

	iMDCT_step_8_1(data_int, N, n4, n8, trig[b][2]);
	iMDCT_step_8_2(data_int, N);
}

#include <cstdio>

void
printData(FLOAT32 *data, int N, const char* fname)
{
#ifdef LINUX_OS
	FILE* fd;

	fd = fopen(fname, "w");
	
	err_printf(fd, "%d\n", N);
	for(int i = 0; i < N; i++) {
		err_printf(fd, "%4.4d: %16.5e\n", i, data[i]);
	}
	fclose(fd);
#endif
}

