
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <math.h>

#ifdef LINUX_OS
#include <asm/byteorder.h>
#include <stdint.h>
#else
extern "C" {
#include "xparameters.h"
#include "mb_interface.h"
}

#endif

#include "my_types.h"
#include "oggStream.h"
#include "codebook.h"
#include "floor1.h"
#include "bitset.h"
#include "residue.h"
#include "vorbis_defs.h"
#include "helpers.h"
#include "overlapper.h"
#include "mdct.h"

#ifdef LINUX_OS

#include "ossAudioPlayer.h"
#include "fileAudioPlayer.h"

#else

#include "AC97Player.h"

#endif

#ifdef HW_OUTPUT
#include "hwOverlapper.h"
#endif

void free_global_data();
int play_ogg_vorbis(OggStream &ogg, int num_packets);
int parse_identity_header(OggStream &ogg);
int parse_comment_header(OggStream &ogg);
int parse_setup_header(OggStream &ogg);
int parse_codebooks(OggStream &ogg);
int parse_time_domain_transforms(OggStream &ogg);
int parse_residues(OggStream &ogg);
int parse_mappings(OggStream &ogg);
int parse_modes(OggStream &ogg);
int prepare_for_decode();

#ifdef LINUX_OS
int
main(int argc, char **argv)
#else
int
main()
#endif
{
	char *filename;

	setbuf(stderr, NULL);
	setbuf(stderr, NULL);

#ifdef LINUX_OS
	if (argc != 2 && argc != 3) {
		printf("Usage: %s <input file> [<num_packets>]\n", argv[0]);
		exit(1);
	}
	filename = argv[1];
	printf("Opening file %s.\n", filename);
#else
//	my_printf("Enabling instruction cache.\n");
//	microblaze_enable_icache();

//	my_printf("Initializing data cache.\n");
//	microblaze_enable_dcache();
#endif

	my_printf("Initializing OggStream\n\r");
	OggStream ogg(filename);
	if (ogg.getStatus() != OggStream::READY) {
		err_printf(stderr, "Failed to initialize ogg stream.\n");
		ogg.printStatus();
		return -1;
	}

	int num_packets = -1;
#ifdef LINUX_OS
	if (argc >= 3) {
		num_packets = atoi(argv[2]);
	}
#endif

	int ret = play_ogg_vorbis(ogg, num_packets);

	free_global_data();

#ifndef LINUX_OS
	my_printf("Entering busy loop.\n");
	while(0);
#endif
	return ret;
}

void
free_global_data()
{
	for(int i = 0; i < 256; i++) {
		free_codebook_data(codebooks[i]);
	}
	for(unsigned i = 0; i < residue_count; i++) {
		clean_residue(residue[i]);
	}
	clean_MDCT();
}

int
play_ogg_vorbis(OggStream &ogg, int num_packets)
{
	int ret = 0;

#ifdef HW_OUTPUT
	HWOverlapper overlapper;
#else
	Overlapper overlapper;
#ifdef LINUX_OS
#ifndef FILE_OUTPUT
	OSSAudioPlayer output;
#else
	FileAudioPlayer output("./output.raw");
#endif // FILE_OUTPUT
#else

	AC97Player output;
	
#endif // LINUX_OS
#endif //  HW_OUTPUT
	// Parse Identity Header
my_printf("Parsing Identity Header\n");
	if ((ret = parse_identity_header(ogg)) < 0) {
		err_printf(stderr, "Error reading Identity Header.\n");
		return ret;
	}

	// Do this here so we'll hear the difference between the init and
	// the first audio packet
my_printf("Initializing Overlapper\n");
	ret = 0;
	ret = overlapper.init(id_header);
	if (ret < 0) {
		err_printf(stderr, "Failed to initialize overlapper.\n");
		return ret;
	}
#ifndef HW_OUTPUT
my_printf("Initializing AC97\n");
	ret = output.init(id_header);
	if (ret < 0) {
		err_printf(stderr, "Failed to initialize Audio Player.\n");
		return ret;
	}
#endif

	// Parse Comment Header
my_printf("Parsing Comment Header\n");
	if ((ret = parse_comment_header(ogg)) < 0) {
		err_printf(stderr, "Error reading Comment Header.\n");
		return ret;
	}

	// Parse Setup header
my_printf("Parsing Setup Header\n");
	if ((ret = parse_setup_header(ogg)) < 0) {
		err_printf(stderr, "Error reading Setup Header.\n");
		return ret;
	}

my_printf("Preparing for decode Header\n");
	if ((ret = prepare_for_decode()) < 0) {
		err_printf(stderr, "Error preparing for decode.\n");
		return ret;
	}

	// Okay, at this point, we're ready to start decoding and playing
	// audio packets.

	int packet = -1;
	do {
		int n_bits = 0;
		int c_bits = 0;
		packet++;
		if (num_packets >= 0) {
			if (packet >= num_packets) {
				break;
			}
		}
my_printf("Decoding packet %d\n", packet);
		FIXINT32** frame = overlapper.nextFrame();
		int size = decode_audio_packet(ogg, frame, c_bits, n_bits);
		if (size < 0) {
			err_printf(stderr, "Error decoding audio packet.\n");
			return ret;
		}
//my_printf("Applying overlap.\n");
#ifdef HW_OUTPUT
		ret = overlapper.applyOverlap(size, c_bits, n_bits);
#else
		ret = overlapper.applyOverlap(size);
#endif
		if (ret < 0) {
			err_printf(stderr, "Error overlapping frames.\n");
			return ret;
		}
#ifndef HW_OUTPUT
//err_printf(stderr, "packet %d in buffer at %p, size = %d (%d)\n", 
//packet, frame, size, overlapper.readyFrameSize());
my_printf("Playing packet.\n");
		ret = output.play(overlapper.readyFrame(),
				(int)overlapper.readyFrameSize());
		if (ret < 0) {
			err_printf(stderr, "Error playing frame.\n");
			return ret;
		}
#endif 
		ogg.advancePacket();
	} while (ogg.getStatus() == OggStream::READY);

	if (ogg.getStatus() != OggStream::EOS) {
		err_printf(stderr, "Did not reach end of Ogg Stream.\n");
		return ret;
	}
	return 0;
}

int
parse_identity_header(OggStream &ogg)
{
	INT32 value;

	my_printf("Parsing Identity Header.\n\r");
	// Make sure it's the right type
	value = ogg.readBits(8);
	if (value != IDENTITY_HEADER_TYPE) {
		err_printf(stderr, "Incorrect header type for identity header.  "
			"Read 0x%hhx, expected 0x%hhx\n",
			value, IDENTITY_HEADER_TYPE);
		return -1;
	}
	// Make sure it's a vorbis header packet
	if (validate_vorbis_header(ogg) < 0) {
		err_printf(stderr, "The Identity header is not a valid vobis "
				"header packet.\n");
		return -1;
	}

	INT32 value1, value2;
	id_header.version = ogg.readBits(32);
	id_header.channels = ogg.readBits(8);
	id_header.sample_rate = value1 = ogg.readBits(32);
	id_header.bitrate_max = ogg.readBits(32);
	id_header.bitrate_nom = value2 = ogg.readBits(32);
	id_header.bitrate_min = ogg.readBits(32);
	id_header.blocksize_0 = 1<<ogg.readBits(4);
	id_header.blocksize_1 = 1<<ogg.readBits(4);
	id_header.framing_flag = ogg.readBits(1);
dbg_printf(stderr, "version %u\n", id_header.version);
dbg_printf(stderr, "channels %hhu\n", id_header.channels);
dbg_printf(stderr, "sample_rate %u %d\n", id_header.sample_rate, value1);
dbg_printf(stderr, "bitrate_max %u\n", id_header.bitrate_max);
dbg_printf(stderr, "bitrate_nom %u %d\n", id_header.bitrate_nom, value2);
dbg_printf(stderr, "bitrate_min %u\n", id_header.bitrate_min);
dbg_printf(stderr, "blocksize_0 %hu\n", id_header.blocksize_0);
dbg_printf(stderr, "blocksize_1 %hu\n", id_header.blocksize_1);

	if (ogg.getStatus() != OggStream::READY) return -1;

	id_header.n_bits[0] = ilog(id_header.blocksize_0/8-1);
	id_header.n_bits[1] = ilog(id_header.blocksize_1/8-1);

	// Check Vorbis Version
	if (id_header.version != 0) {
		err_printf(stderr, "Incompatible vorbis version 0x%x.\n", 
			id_header.version);
		return -1;
	}
	
	// Validate Channels and Sample Rate both gt 0.
	if (id_header.channels == 0) {
		err_printf(stderr, "0 audio channels present. Exiting...\n");
		return -1;
	}
	if (id_header.sample_rate == 0) {
		err_printf(stderr, "No sample rate specified in header.\n");
		return -1;
	}

	// Validate blocksizes
	if (id_header.blocksize_0 > id_header.blocksize_1) {
		err_printf(stderr, "Blocksize_0 is greater than Blocksize_1 "
				"in identity header.  Invalid.\n");
		return -1;
	}
	if (id_header.blocksize_0 < 64 || id_header.blocksize_0 > 8192) {
		err_printf(stderr, "Invalid value for blocksize_0.\n");
		return -1;
	}
	if (id_header.blocksize_1 < 64 || id_header.blocksize_1 > 8192) {
		err_printf(stderr, "Invalid value for blocksize_1.\n");
		return -1;
	}

	// Make sure framing flag is set.
	if (id_header.framing_flag != 1) {
		err_printf(stderr, "Error: Framing flag not set in Identity "
				"header.\n");
		return -1;
	}

dbg_printf(stderr, "Sample rate is %u\n", id_header.sample_rate);

	// Okay, at this point we've extracted the header info and validated
	// all the fields, so we can continue
	return ogg.advancePacket();
}

int
parse_comment_header(OggStream &ogg)
{
	INT32 value;
	INT32 length;
	UINT32 num_strs;
	UINT32 i;
	
	
	my_printf("Reading comment header.\n");
	// Make sure it's the right type
	if ((value = ogg.readBits(8)) != COMMENT_HEADER_TYPE) {
		err_printf(stderr, "Incorrect header type for comment header.  "
			"Read 0x%x, expected 0x%hhx\n",
			value, COMMENT_HEADER_TYPE);
		ogg.printStatus();
		return -1;
	}
	// Make sure it's a vorbis header packet
	if (validate_vorbis_header(ogg) < 0) {
		err_printf(stderr, "The Comment header is not a valid vobis "
				"header packet.\n");
		return -1;
	}

	// Print the vendor string
	length = (UINT32)ogg.readBits(32);
//dbg_printf(stderr, "reading %u octects\n", length);
	err_printf(stderr, "Vendor String: ");
	for(int i = 0; i < length; i++) {
		err_printf(stderr, "%c", ogg.readBits(8));
	}
	err_printf(stderr, "\n");

	// Print all the comments strings
	num_strs = ogg.readBits(32);
//dbg_printf(stderr, "reading %u strings\n", num_strs);
	for(i = 0; i < num_strs; i++) {
		length = ogg.readBits(32);
//dbg_printf(stderr, "reading %u octects\n", length);
#ifdef LINUX_OS
		fprintf(stderr,"Comment %d: ", i);
		for(int i = 0; i < length; i++) {
			fprintf(stderr, "%c", ogg.readBits(8));
		}
		fprintf(stderr, "\n");
#else
		xil_printf("Comment %d: ", i);
		for(int i = 0; i < length; i++) {
			xil_printf("%c", ogg.readBits(8));
		}
		xil_printf("\n\r");
#endif
		if (ogg.getStatus() != OggStream::READY) {
			return -1;
		}
	}

	// Check Framing bit
	if (ogg.readBits(1) != 1) {
		err_printf(stderr, "Framing bit not set in comment header.\n");
		return -1;
	}

	return ogg.advancePacket();
}

int
parse_setup_header(OggStream &ogg)
{
	int ret = 0;
	INT32 value;

	my_printf("Parsing setup header.\n");
	// Make sure it's the right type
	if ((value = ogg.readBits(8)) != SETUP_HEADER_TYPE) {
		err_printf(stderr, "Incorrect header type for setup header.  "
			"Read 0x%hhx, expected 0x%hhx\n",
			value, SETUP_HEADER_TYPE);
		return -1;
	}
	// Make sure it's a vorbis header packet
	if (validate_vorbis_header(ogg) < 0) {
		err_printf(stderr, "The Setup header is not a valid vobis "
				"header packet.\n");
		return -1;
	}

	if (ogg.getStatus() != OggStream::READY) return -1;

	// Parse and decode the codebooks
	if ((ret = parse_codebooks(ogg)) < 0) {
		return ret;
	}

	// Parse and decode the time domain tranforms
	if ((ret = parse_time_domain_transforms(ogg)) < 0) {
		return ret;
	}

	// Parse and decode the floors
	if ((ret = parse_floors(ogg)) < 0) {
		return ret;
	}

	// Parse and decode the residues
	if ((ret = parse_residues(ogg)) < 0) {
		return ret;
	}

	// Parse and decode the mappings
	if ((ret = parse_mappings(ogg)) < 0) {
		return ret;
	}

	// Parse and decode the modes
	if ((ret = parse_modes(ogg)) < 0) {
		return ret;
	}

	if (ogg.readBits(1) != 1) {
		err_printf(stderr, "Framing error while reading setup header.  "
				"Undecodable stream.\n");
		ret = -1;
	}

	return ogg.advancePacket();
}

int
parse_codebooks(OggStream &ogg)
{
	UINT16 i;
	int ret = 0;

	codebook_count = ogg.readBits(8);
	if (ogg.getStatus() != OggStream::READY) return -1;
	codebook_count++;

	// Once we start reading the codebooks, there are no guarantees that
	// the fields will be aligned on any kind of boundaries, so we have
	// to read everything bit by bit.
	for(i = 0; i < codebook_count; i++) {
//dbg_printf(stderr, "Decoding codebool %d of %d\n", i+1, codebook_count);
		UINT8 sync[3];
		UINT32 j = 0;
		codebook_t *cb = &(codebooks[i]);
		//First, check the sync pattern
		sync[0] = ogg.readBits(8);
		sync[1] = ogg.readBits(8);
		sync[2] = ogg.readBits(8);
		if (sync[0] != 0x42 || sync[1] != 0x43 || sync[2] != 0x56) {
			err_printf(stderr, "Invalid sync pattern found in "
					"codebook %hhu (%hhx:%hhx:%hhx).\n", 
					i, sync[0], sync[1], sync[2]);
			return -1;
		}
		// Now extract the dimensions field
		cb->dimensions = ogg.readBits(16);
		// Get the number of entries
		cb->num_entries = ogg.readBits(24);
		cb->ordered = ogg.readBits(1);

		cb->lengths = (UINT8*)my_alloc(cb->num_entries);
		if (cb->ordered) {
			UINT8 len = ogg.readBits(5);
			for(j = 0; j < cb->num_entries; j++) {
				UINT8 number;
				UINT8 limit;
				number = ogg.readBits(ilog(cb->num_entries-j));
				limit = j + number;
				for(;j < cb->num_entries && j < limit;j++) {
					cb->lengths[j] = len;
				}
				len++;
				if (j > cb->num_entries) {
					err_printf(stderr, "Error Condition! "
						"Error decoding codebook "
						"lengths.\n");
					return -1;
				}
			}
		} else {	// Not ordered
			cb->sparse = ogg.readBits(1);
			for(j = 0; j < cb->num_entries; j++) {
				UINT8 len;
				if (cb->sparse) {
					UINT8 flag = ogg.readBits(1);
					if (flag) {
						len = ogg.readBits(5);
						cb->lengths[j] = len+1;
					} else {
						cb->lengths[j] = 0;
					}
				} else {
					len = ogg.readBits(5);
					cb->lengths[j] = len+1;
				}
			}
		}

		// Now read the codebook lookup type
		cb->lookup_type = ogg.readBits(4);

		// Read lookup values if necessary
		if (cb->lookup_type == 1 || cb->lookup_type == 2) {
			UINT32 k;
			cb->min_val = float32_unpack(ogg.readBits(32));
			cb->delta_val = float32_unpack(ogg.readBits(32));
			cb->value_bits = (UINT8)(ogg.readBits(4) & 0xf) + 1;
			cb->sequence_p = ogg.readBits(1);
			if (cb->lookup_type == 1) {
				cb->lookup_values = lookup1_values(
						cb->num_entries, 
						cb->dimensions);
			} else {	// cb->lookup_type == 2
				cb->lookup_values = cb->num_entries *
							cb->dimensions;
			}
			cb->multiplicands = (UINT16*)my_alloc(
					cb->lookup_values * sizeof(UINT16));
			for(k = 0; k < cb->lookup_values; k++) {
				cb->multiplicands[k] =
					ogg.readBits(cb->value_bits);
			}
		} else if (cb->lookup_type != 0) {
			err_printf(stderr, "Lookup type %hhu unknown.\n",
					cb->lookup_type);
			return -1;
		}

		// Decode the VQ lookup table
		ret = decode_lookup_table(cb);
		if (ret < 0) {
			err_printf(stderr, "Failed to decode lookup table for "
					"codebook %hhu.\n", i);
			return ret;
		}

		// Build a huffman decision tree for this codebook
		ret = build_huffman_tree(cb);
		if (ret < 0) {
			err_printf(stderr, "Failed to build huffman tree for "
					"codebook %u.\n", i);
			return ret;
		}
	}

	return 0;
}

int
parse_time_domain_transforms(OggStream &ogg)
{
	UINT8 time_count;
	UINT8 i;

	time_count = ogg.readBits(6);
	if (ogg.getStatus() != OggStream::READY) return -1;
	time_count++;
	for(i = 0; i < time_count; i++) {
		INT32 temp;
		if ((temp = ogg.readBits(16)) != 0) {
			err_printf(stderr, "Non-zero value (0x%04x) while parsing"
					" time domain transform %hhu/%hhu.  "
					"Undecodeable stream.\n", temp, i,
					time_count);
			return -1;
		}
	}

	return 0;
}

int
parse_residues(OggStream &ogg)
{
	UINT8 i;

	residue_count = ogg.readBits(6);
	residue_count++;
	for(i = 0; i < residue_count; i++) {
dbg_printf(stderr, "Decoding residue %d\n", i);
		UINT8 j;
		residue[i].type = ogg.readBits(16);
		if (residue[i].type > 2) {
			err_printf(stderr, "Invalid residue type %hu. "
					"Undecodable stream.\n", 
					residue[i].type);
			return -1;
		}
		residue[i].begin = ogg.readBits(24);
		residue[i].end = ogg.readBits(24);
		residue[i].partition_size = ogg.readBits(24)+1;
		residue[i].classifications = ogg.readBits(6)+1;
		residue[i].classbook = ogg.readBits(8);
		for(j = 0; j < residue[i].classifications; j++) {
			UINT8 bit_flag, low_bits, high_bits = 0;
			low_bits = ogg.readBits(3);
			bit_flag = ogg.readBits(1);
			if (bit_flag) {
				high_bits = ogg.readBits(5);
			}
			residue[i].cascade[j] = (high_bits<<3)|(0x7&low_bits);
//dbg_printf(stderr, "Residue %d cascade %d = %d, (hb=%hhu, lb=%hhu, bf=%hhu)\n", 
//i, j, residue[i].cascade[j], high_bits, low_bits, bit_flag);
		}
		for(j = 0; j < residue[i].classifications; j++) {
			UINT8 k, mask;
			for(k = 0, mask = 1; k < 8; k++, mask<<=1) {
				if ((residue[i].cascade[j] & mask) != 0) {
					residue[i].books[j][k] = 
							ogg.readBits(8);
//dbg_printf(stderr, "Residue %d, book[%d][%d] is %d.\n", i, j, k,
//residue[i].books[j][k]);
				} else {
//dbg_printf(stderr, "Residue %d, book[%d][%d] is unused.\n", i, j, k);
					residue[i].books[j][k] = 
						RESIDUE_BOOK_UNUSED;
				}
			}
		}
	}

	return 0;
}

int
parse_mappings(OggStream &ogg)
{
	UINT8 i;

	mapping_count = ogg.readBits(6);
	mapping_count++;
	for(i = 0; i < mapping_count; i++) {
		int value;
		UINT16 j;
		mapping_t *mapping = &(mappings[i]);
		// Read the mapping type and make sure it's zero
		if ((value = ogg.readBits(16)) != 0) {
			err_printf(stderr, "Invalid mapping type %hu.  "
					"Undecodable stream.\n", value);
			return -1;
		}
		if (ogg.readBits(1)) {
			mapping->num_submaps = ogg.readBits(4)+1;
		} else {
			mapping->num_submaps = 1;
		}
		if (ogg.readBits(1)) {
			UINT8 num_bits = ilog(id_header.channels - 1);
			mapping->coupling_steps = ogg.readBits(8) + 1;
			for(j = 0; j < mapping->coupling_steps; j++) {
				mapping->magnitude[j] = ogg.readBits(num_bits);
				mapping->angle[j] = ogg.readBits(num_bits);
				if (mapping->magnitude[j] == mapping->angle[j]){
					err_printf(stderr, "Angle channel is the "
							"same as magnitude "
							"channel.  Undecodable "
							"stream.\n");
					return -1;
				} else if (mapping->magnitude[j] >=
						id_header.channels) {
					err_printf(stderr, "Invalid channel number"
							" for magnitude mapping"
							".\n");
					return -1;
				} else if (mapping->angle[j] >=
						id_header.channels) {
					err_printf(stderr, "Invalid channel number"
							" for angle mapping"
							".\n");
					return -1;
				} 
			}
		} else {
			mapping->coupling_steps = 0;
		}
		if (ogg.readBits(2) != 0) {
			err_printf(stderr, "Non-zero value in reserved field.  "
					"Undecodable stream.\n");
			return -1;
		}
		if (mapping->num_submaps > 1) {
			for(j = 0; j < id_header.channels; j++) {
				mapping->mux[j] = ogg.readBits(4);
				if (mapping->mux[j] >= mapping->num_submaps) {
					err_printf(stderr, "Error, mapping_mux "
							"value higher than "
							"highest submap index"
							".\n");
					return -1;
				}
			}
		}
		for(j = 0; j < mapping->num_submaps; j++) {
			// Read and discard time config placeholder
			if (ogg.readBits(8) < 0) return -1;
			// Get floor number
			mapping->submap[j].floor = ogg.readBits(8);
			if (mapping->submap[j].floor >= floor_count) {
				err_printf(stderr, "Invalid floor number. "
						"Undecodable stream.\n");
				return -1;
			}
			// Get residue number
			mapping->submap[j].residue = ogg.readBits(8);
			if (mapping->submap[j].residue >= residue_count) {
				err_printf(stderr, "Invalid residue number. "
						"Undecodable stream.\n");
				return -1;
			}
		}
	}

	return 0;
}

int
parse_modes(OggStream &ogg)
{
	UINT8 i;

	mode_count = ogg.readBits(6) + 1;
dbg_printf(stderr, "Reading in %u modes.\n", mode_count);
	for(i = 0; i < mapping_count; i++) {
		mode_config_t *mode = &(modes[i]);
dbg_printf(stderr, "Reading mode %hhu blockflag at ", i);
ogg.printPosition();
		mode->blockflag = ogg.readBits(1);
dbg_printf(stderr, " bit is %hhu\n", mode->blockflag);
		// Read window type and verify it's 0
		if (ogg.readBits(16) != 0) {
			err_printf(stderr, "Mode window type is nonzero.  "
					"Undecodable stream.\n");
			return -1;
		}
		// Read transform type and verify it's 0
		if (ogg.readBits(16) != 0) {
			err_printf(stderr, "Mode transform type is nonzero.  "
					"Undecodable stream.\n");
			return -1;
		}
		if ((mode->mapping = ogg.readBits(8)) >= mapping_count) {
			err_printf(stderr, "Mode mapping is outside valid range.  "
					"Undecodable stream.\n");
			return -1;
		}
	}

	return 0;
}

int
prepare_for_decode()
{
	init_MDCT(id_header.blocksize_0, id_header.blocksize_1);
	for(unsigned i = 0; i < residue_count; i++) {
		prepare_residue(residue[i]);
	}
	return 0;
}
