#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "my_types.h"
#include "oggStream.h"
#include "codebook.h"
#include "floor1.h"
#include "bitset.h"
#include "residue.h"
#include "vorbis_defs.h"
#include "helpers.h"
#include "overlapper.h"

Overlapper::~Overlapper()
{
	if (ready_buf != NULL) {
		for(int i = 0; i < channels; i++) {
			if (ready_buf[i] != NULL) free(ready_buf[i]);
		}
		free(ready_buf);
		ready_buf = NULL;
	}
	if (buffer[0] != NULL) {
		for(int i = 0; i < channels; i++) {
			if (buffer[0][i] != NULL) free(buffer[0][i]);
		}
		free(buffer[0]);
		buffer[0] = NULL;
	}
	if (buffer[1] != NULL) {
		for(int i = 0; i < channels; i++) {
			if (buffer[1][i] != NULL) free(buffer[1][i]);
		}
		free(buffer[1]);
		buffer[1] = NULL;
	}
}

int
Overlapper::init(id_header_t &header)
{
	blocksize0 = header.blocksize_0;
	blocksize1 = header.blocksize_1;
	channels = header.channels;

	UINT16 max_blocksize = blocksize1;

	// We should have already check for valid blocksizes,
	// but we really want to be sure the buffers are big enough
	if (blocksize0 > max_blocksize) {
		max_blocksize = blocksize0;
	}

	buffer[0] = (FIXINT32**)my_alloc(channels*sizeof(FIXINT32*));
	if (buffer[0] == NULL) {
		return -1;
	}
	buffer[1] = (FIXINT32**)my_alloc(channels*sizeof(FIXINT32*));
	if (buffer[1] == NULL) {
		free(buffer[0]);
		buffer[0] = NULL;
		return -1;
	}
	ready_buf = (INT16**)my_alloc(channels*sizeof(INT16*));
	if (ready_buf == NULL) {
		free(ready_buf);
		ready_buf = NULL;
		return -1;
	}
	// Initialize buffers to zero so they can be properly freed on error
	for(int i = 0; i < channels; i++) {
		buffer[0][i] = NULL;
		buffer[1][i] = NULL;
		ready_buf[i] = NULL;
	}
	for(int i = 0; i < channels; i++) {
		for(int j = 0; j < 2; j++) {
			buffer[j][i] = (FIXINT32*)my_alloc(max_blocksize
						*sizeof(FIXINT32));
			if (buffer[j][i] == NULL) {
				goto init_err;
			}
		}
		ready_buf[i] = (INT16*)my_alloc((max_blocksize/2)*sizeof(INT16));
		if (ready_buf[i] == NULL) {
			goto init_err;
		}
	}
	
	last_buffer = 1;
	ready_buffer = -2;

	return 0;

init_err:
	if (ready_buf != NULL) {
		for(int i = 0; i < channels; i++) {
			if (ready_buf[i] != NULL) free(ready_buf[i]);
		}
		free(ready_buf);
		ready_buf = NULL;
	}
	if (buffer[0] != NULL) {
		for(int i = 0; i < channels; i++) {
			if (buffer[0][i] != NULL) free(buffer[0][i]);
		}
		free(buffer[0]);
		buffer[0] = NULL;
	}
	if (buffer[1] != NULL) {
		for(int i = 0; i < channels; i++) {
			if (buffer[1][i] != NULL) free(buffer[1][i]);
		}
		free(buffer[1]);
		buffer[1] = NULL;
	}
	return -1;
}

static inline INT16
float32_to_int16(FLOAT32 f_val)
{
	INT32 val = (INT32)(f_val * 32767.0f);

	// Check for clipping
	if (val > 32767) {
		val = 32767;
	} else if (val < -32768) {
		val = -32768;
	}
	return (INT16)val;
}

int 
Overlapper::applyOverlap(int blocksize)
{
	if (last_buffer < 0) {
		return -1;
	}
	if (ready_buffer == -2) {
		ready_buffer = -1;
		last_blocksize = blocksize;
		return 0;
	} else if (ready_buffer == -1) {
		ready_buffer = 0;
	}

	int p_n = last_blocksize;
	int c_n = blocksize;
	int win_start = 0;
	int win_end = c_n / 2;
	int win_n = c_n / 2;
	if (c_n > p_n) {
		win_start = (c_n/4) - (p_n/4);
		win_end = (c_n/4) + (p_n/4);
		win_n = p_n/2;
	}
	// Apply the overlap;
	for(int i = 0; i < channels; i++) {
		FIXINT32* prev = buffer[last_buffer^0x1][i];
		FIXINT32* cur = buffer[last_buffer][i];
		INT16* out = ready_buf[i];

		prev += p_n/2;
		cur  += win_start;

		for(int i = 0; i < win_start; i++,prev++,out++) {
			*out = prev->int16();
		}
		
		for(int i = 0; i < win_n; i++,out++,prev++,cur++) {
			FLOAT32 a = sin(0.5*pi*(i+0.5)/(FLOAT32)win_n);
			a *= a;
			FIXINT32 val = (*cur) * FIXINT32((float)sin(0.5*a*pi));

			a = sin(0.5*pi*(win_n - i - 0.5)/(FLOAT32)win_n);
			a *= a;
			val += (*prev) * FIXINT32((float)sin(0.5*a*pi));
			*out = val.int16();
		}
		for(int i = win_n; i < win_end; i++,out++,cur++) {
			*out = cur->int16();
		}
	}
	
	ready_blocksize = (c_n/4) + (p_n/4);
	last_blocksize = blocksize;
	return 0;
}

INT16**
Overlapper::readyFrame()
{
	if (ready_buffer < 0) {
		return NULL;
	} else {
		return ready_buf;
	}
}

FIXINT32**
Overlapper::nextFrame()
{
	last_buffer ^= 0x1;
	return buffer[last_buffer];
}

