#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "my_types.h"
#include "oggStream.h"
#include "codebook.h"
#include "floor1.h"
#include "bitset.h"
#include "residue.h"
#include "helpers.h"
#include "mdct.h"
#include "vorbis_defs.h"

id_header_t	id_header;
UINT16		codebook_count;
UINT8		floor_count;
UINT8		residue_count;
UINT8		mapping_count;
UINT8		mode_count;
codebook_t	codebooks[256];
floor_config_t	floor_config[64];
residue_t	residue[64];
mapping_t	mappings[64];
mode_config_t	modes[64];
Bitset		no_residue(256);
Bitset		do_not_decode(256);

int
validate_vorbis_header(OggStream &ogg)
{
	// Make sure it's a vorbis header packet
	if (		ogg.readBits(8) != 'v' ||
			ogg.readBits(8) != 'o' ||
			ogg.readBits(8) != 'r' ||
			ogg.readBits(8) != 'b' ||
			ogg.readBits(8) != 'i' ||
			ogg.readBits(8) != 's'	) {
		return -1;
	}
	return 0;
}

void
printFloorYVals(INT32 *vals, int num)
{
#ifdef LINUX_OS
fprintf(stderr, "Printing floor y values.\n");
	FILE* fd = fopen("floor.out", "w");
	fprintf(fd, "%d\n", num);
	for(int i = 0; i < num; i++) {
		fprintf(fd, "%d\n", vals[i]);
	}
	fclose(fd);
#endif
}

void
printResidue(FIXINT32 *vals, int num, const char* fname)
{
#ifdef LINUX_OS
fprintf(stderr, "Printing Residue values.\n");
	FILE* fd = fopen(fname, "w");
	fprintf(fd, "%d\n", num);
	for(int i = 0; i < num; i++) {
		fprintf(fd, "%16e\n", vals[i].getFloat());
	}
	fclose(fd);
#endif
}

int
decode_audio_packet(OggStream &ogg, FIXINT32** pcm, int &c_bits, int &n_bits)
{
	static int count = 0;
	mode_config_t *mode = NULL;
	mapping_t *mapping = NULL;
	INT32 value, ret = 0;
	UINT16 N;
	UINT8 mode_number, i;
	UINT8 next_window, prev_window;
	INT32 **floor_y_vals = NULL;
	FIXINT32 *pcmbundle[2] = { NULL, NULL};

dbg_printf(stderr, "Decoding audio packet %d starting at ", ++count);
ogg.printPosition();
dbg_printf(stderr, "\n");
//#ifdef VERBOSE_OUTPUT
//if (count > 12) return -1;
//#endif
	if (ogg.readBits(1) != 0) {
		fprintf(stderr, "decode_audio_packet: packet is not an "
				"audio packet.\n");
		return -1;
	}
dbg_printf(stderr, "Reading %u bits for mode_number.\n", ilog(mode_count-1));
	value = ogg.readBits(ilog(mode_count-1));
	if (value < 0) {
		ret = value;
		goto ogg_error;
	}
	mode_number = value;
	if (mode_number >= mode_count) {
		fprintf(stderr, "Invalid mode number %hhu\n", mode_number);
		goto ogg_error;
		return -1;
	}
	mode = &(modes[mode_number]);
	mapping = &(mappings[mode->mapping]);
	c_bits = id_header.n_bits[mode->blockflag];
	if (mode->blockflag == 1) {
dbg_printf(stderr, "Mode %hhu has a Long window, reading 2 more bits.\n", 
mode_number);
		N = id_header.blocksize_1;
		
		value = ogg.readBits(1);
		if (value < 0) goto ogg_error;
		prev_window = (UINT8)value;
		
		value = ogg.readBits(1);
		if (value < 0) goto ogg_error;
		next_window = (UINT8)value;
		n_bits = id_header.n_bits[next_window];
	} else {
dbg_printf(stderr, "Mode %hhu has a Short window.\n", 
mode_number);
		N = id_header.blocksize_0;
		n_bits = c_bits;
	}
	
	// Zero pcm data
	for(int i = 0; i < id_header.channels; i++) {
		bzero((char*)pcm[i], N*sizeof(FIXINT32));
	}

	// Allocate floor array
	floor_y_vals = (INT32**)my_alloc(id_header.channels*sizeof(INT32));
	bzero((char*)floor_y_vals, id_header.channels*sizeof(INT32*));

	// Past this point, it's acceptable for ogg.readBits() to
	// fail with status EOP

	no_residue.setAllBits(false);
	for(i = 0; i < id_header.channels; i++) {
dbg_printf(stderr, "decoding floor for channel %d at", i);
ogg.printPosition();
		UINT8 submap = mapping->mux[i];
		UINT32 f = mapping->submap[submap].floor;
		floor_config_t &floor = floor_config[f];
		floor_y_vals[i] = NULL;
		floor_y_vals[i] = (INT32*)my_alloc(floor.num_positions*
					sizeof(INT32));
		assert(floor_y_vals[i] != NULL);
		bzero((char*)floor_y_vals[i], floor.num_positions*sizeof(INT32));
dbg_printf(stderr, "allocated floor %hhu at %p (size = %u)\n", 
i, floor_y_vals[i], floor.num_positions);
		// The floor must be type 1 (we'd have died already ow)
		
		// Decode Floor1 according to section 7.2.2.1 "packet decode"
		ret = decode_floor1_packet(ogg, floor, floor_y_vals[i]);
if (count == 12 && i == 0)
printFloorYVals(floor_y_vals[i], floor.num_positions);
		if (ret < 0) {
			fprintf(stderr, "Failed to decode floor1 packet.\n");
			goto ogg_error;
		}
		if (ret == FLOOR_UNUSED) {
dbg_printf(stderr, "residue %d is UNUSED.\n", i);
			no_residue.setBit(i, true);
		}
		if (ogg.getStatus() == OggStream::EOP) {
			ret = N;
			goto eop_out;
		}
	}
dbg_printf(stderr, "After decoding floors, stream is at position: ");
ogg.printPosition();
//ret = 0;
//goto ogg_error;

	// Make sure that coupled channels are not zeroed because one is unused
	for(UINT16 i = 0; i < mapping->coupling_steps; i++) {
		if (!no_residue.getBit(mapping->magnitude[i])) {
			no_residue.setBit(mapping->angle[i], false);
		}
		if (!no_residue.getBit(mapping->angle[i])) {
			no_residue.setBit(mapping->magnitude[i], false);
		}
	}
	
	// Decode the Residue Vectors
	//pcmbundle = (FIXINT32**)my_alloc( id_header.channels*sizeof(FIXINT32*));
	// Use a static array of size 2 to avoid malloc/free calls
	if (id_header.channels > 2) {
		err_printf(stderr, "Too many channels.\n");
		ret = -1;
		goto ogg_error;
	}

	for(UINT8 i = 0; i < mapping->num_submaps; i++) {
//dbg_printf(stderr, "no_residue bitmap = ");
//no_residue.print();
//dbg_printf(stderr, "\n");
		int ch = 0;
		do_not_decode.setAllBits(true);
		for(UINT8 j = 0; j < id_header.channels; j++) {
dbg_printf(stderr, "mux[%hhu] = %hhu (want == %hhu).\n", j, mapping->mux[j], i);
			if (mapping->mux[j] == i) {
dbg_printf(stderr, "%sSetting bit %d\n", (no_residue.getBit(j) ? "":"UN-"), ch);
				do_not_decode.setBit(ch, no_residue.getBit(j));
dbg_printf(stderr, "bit %d is %s\n", ch, 
  (do_not_decode.getBit(ch) ? "SET" : "NOT set"));
dbg_printf(stderr, "pcmbundle[%d] = pcm[%d] = %p\n", ch, j, pcm[j]);
			  //build a mirrors PCM buddle (only needed temporarily
			  //**TODO** TAKE THIS OUT FOR SURE VERY VERY SLOW
			/* pcmbundle[ch] = (FIXINT32*)my_alloc(sizeof(FIXINT32)*N);
				for (int k = 0; k < N; k ++)
					pcmbundle[ch][k] = pcm[j][k];
				ch++;*/
				pcmbundle[ch++] = pcm[j];
			}
		}
		UINT8 residue_number = mapping->submap[i].residue;

dbg_printf(stderr, "Decoding residue for %d channel(s) using res %d\n", ch,
residue_number);
		ret = decode_residue(ogg, ch, N/2, residue[residue_number],
					pcmbundle, do_not_decode);
					
		//TODO ** REMOVE THIS SILLY COPY BACK
		/*for (int j = 0 ; j <id_header.channels; j++){
			for (int k = 0; k < N; k++)
					pcm[j][k] = pcmbundle[j][k].getFloat();
		}*/
		
if (count == 12 && i == 0)
printResidue(pcm[0], N/2, "residue.out");
		if (ret < 0) {
			fprintf(stderr, "decode_residue failed.\n");
			goto ogg_error;
		}
	}
	//free(pcmbundle);
	//pcmbundle = NULL;

dbg_printf(stderr, "After decoding residues, stream is at position: ");
ogg.printPosition();
	// Inverse Coupling
	for(UINT16 i = 0; i < mapping->coupling_steps; i++) {
		FIXINT32* mag = pcm[mapping->magnitude[i]];
		FIXINT32* ang = pcm[mapping->angle[i]];
		
		for(int j = 0; j < N/2; j++) {
			if (mag[j] > 0) {
				if (ang[j] <= 0) {
					mag[j] += ang[j];
				}
				ang[j] = mag[j] - ang[j];
			} else {
				if (ang[j] <= 0) {
					mag[j] -= ang[j];
				}
				ang[j] += mag[j];
			}
		}
	}
if (count == 12)
printResidue(pcm[0], N/2, "uncoupled.out");

	// Synthesize floor curve and compute dot product with residue
	// And while we're looping through the channels, do the iMDCT too
	for(UINT8 i = 0; i < id_header.channels; i++) {
		UINT8 floor = mapping->submap[mapping->mux[i]].floor;
		
		floor1_synthesis(pcm[i], N/2, floor_y_vals[i], 
				floor_config[floor]);
		
		inverse_MDCT(pcm[i], N, c_bits);
	}

	if (floor_y_vals != NULL) {
		for(UINT8 i = 0; i < id_header.channels; i++) {
			if (floor_y_vals[i] != NULL) {
dbg_printf(stderr, "free floor %hhu at %p\n", i, floor_y_vals[i]);
				free(floor_y_vals[i]);
			}
		}
		free(floor_y_vals);
	}

	return (int)N;

eop_out:
ogg_error:
	//if (pcmbundle != NULL) free(pcmbundle);
	if (floor_y_vals != NULL) {
		for(UINT8 i = 0; i < id_header.channels; i++) {
			if (floor_y_vals[i] != 0) {
dbg_printf(stderr, "free floor %hhu at %p\n", i, floor_y_vals[i]);
				free(floor_y_vals[i]);
			}
		}
		free(floor_y_vals);
	}
	ogg.printStatus();
	return ret;
}
